"""
    Definition of the backward-compatible CLI interface of PyPhoneBook.
"""
from typing import Any, Dict, List, Tuple
import sys
import traceback
from textwrap import dedent
from argparse import ArgumentParser, SUPPRESS, RawTextHelpFormatter
from pyphonebook import PhoneBook, PhoneBookEntry, PHONEBOOK_AND, PHONEBOOK_OR, _TERSE_ALIASES


_ORDERBY_ALIASES = {"surname": "last_name",
                   "firstname": "first_name",
                   "login": "login_name",
                   "department": "department",
                   "group": "_extensionAttribute13",
                   "cerngroup": "group",
                   "cernsection": "section",
                   "displayname": "_displayName"}


def _cli() -> None:
    """
    Defines the CLI interface with ``argparse`` and uses
    ``PhoneBook.compound_search()`` to print the results.

    Output formatting is not done here: see PhoneBookEntry.__str__()

    Can be used a simple usage example.
    """
    verbose = True  # In case it fails before reading the flag
    try:

        # Create the parser for CLI arguments and define the expected flags
        parser = ArgumentParser(formatter_class=RawTextHelpFormatter,
                                description=dedent("""\
                                phonebook - Utility to search CERN phone book.


                                'phonebook' is a replacement for a replacement for an aging phone command well known
                                to all CERN unix users. It will search CERN LDAP database for user information and
                                display the output according to search criteria defined."""),

                                epilog=dedent("""\
                                All options can be abbreviated to shortest distinctive length, (first letter in most
                                cases). Single minus preceding option name may be used instead of double one.

                                For wildcard search use * character (double quoted or back-slashed to avoid shell
                                variable expansion: \"*\" or \\* ).

                                Please note that search option arguments are case insensitive.
                                The number of returned results is limited to 1000."""))

        parser.add_argument('search_string', nargs="?", metavar="SEARCHSTRING", default=SUPPRESS)

        parser.add_argument('--surname', metavar="SURNAME", dest="last_name", help="search by last name",
                            default=SUPPRESS)

        parser.add_argument('--firstname', metavar="FIRSTNAME", dest="first_name", help="search by first name",
                            default=SUPPRESS)

        parser.add_argument('--building', metavar="BUILDING", help="search by building", default=SUPPRESS)

        parser.add_argument('--office', metavar="FLOOR-OFFICE | OFFICE | FLOOR", help="search by office number",
                            default=SUPPRESS)

        parser.add_argument('--email', metavar="EMAIL@ADDRESS", help="search by email", default=SUPPRESS)

        parser.add_argument('--login', metavar="LOGIN", dest="login_name", help="search by login name", default=SUPPRESS)

        parser.add_argument('--department', metavar="DEPARTMENT", help="search by department code (two-letter code)",
                            default=SUPPRESS)

        parser.add_argument('--group',  metavar="GROUP", help="search by group code (two-letter code",
                            default=SUPPRESS)

        parser.add_argument('--section', metavar="SECTION", help="search by section code (two-letter code)",
                            default=SUPPRESS)

        parser.add_argument('--phone', metavar="PHONE", dest="landline_phone", help="search by phone number",
                            default=SUPPRESS)

        parser.add_argument('--mobile', metavar="MOBILE", dest="mobile_phone", help="search by phone number",
                            default=SUPPRESS)

        parser.add_argument('--pobox', metavar="POBOX", dest="post_box", help="search by post box", default=SUPPRESS)

        parser.add_argument('--ccid', metavar="COMPUTER-CENTER-ID", help="search by Computing Center ID",
                            default=SUPPRESS)

        parser.add_argument('--uid', metavar="UID", help="search by UID", default=SUPPRESS)

        parser.add_argument('--gid', metavar="GID", help="search by GID", default=SUPPRESS)

        parser.add_argument('--homedir', metavar="HOMEDIR", dest="home_directory", help="search by home directory",
                            default=SUPPRESS)

        # Filtering, Ordering & Formatting
        domain_mutexes = parser.add_mutually_exclusive_group()
        formatting_mutexes = parser.add_mutually_exclusive_group()

        domain_mutexes.add_argument('--externals', action="store_true", default=SUPPRESS,
                                    help="search only for people having external / lightweight accounts registered.")

        domain_mutexes.add_argument('--yellowpages', action="store_true", default=SUPPRESS,
                                    help="search for services instead of people.")

        formatting_mutexes.add_argument('--all', action="store_true", default=SUPPRESS,
                            help="show all fields in the output.")

        parser.add_argument('--orderby', metavar="OOO", action="append", default=SUPPRESS,
                            help=dedent("""\
                                 order the output sorting (alphanumerically) values of one of the fields.
                                 Valid OOO values are: surname, firstname, login, department, group, cerngroup,
                                 cernsection or displayname."""))

        formatting_mutexes.add_argument('--terse', metavar="TTT", action="append", default=SUPPRESS,
                            help=dedent("""\
                                     show only TTT fields in the output. This output option allows to fully
                                     customize what phonebook prints in the output listing.

                                     Terse option can be specified multiple times on the command line.
                                     Output values are listed in same order as input terse options and are
                                     separated by ';' sign. Multiple-value values are separated by ','.
                                     Empty values are printed as '-'.

                                     Names of most of the terse option arguments are self-explanatory, except:

                                     uac - user account control: this field returns value of UAC field in
                                     Active Directory in decimal and permits to check if the account is
                                     blocked and for what reason: to decode the value please see:
                                     http://support.microsoft.com/kb/305144

                                         type - account type: value can be P - for primary account, S - for
                                         secondary account, s - for service account

                                         secid - alternate security identities mapped to the account (ssh keys,
                                         X509 certificates, YubiKey(s) .. etc)

                                            Note: identifies are printed out as:

                                            '[TYPE:Content]' where TYPE can be SSH,X509,YubiKey
                                                             and Content value depends on type."""))

        parser.add_argument("-v", "--verbose", action="store_true",
                            help="Prints error messages if they arise during parsing")

        # Remove the default -h flag for help because the damn Perl version uses it for searching home directories...
        parser._actions[0].option_strings = [opt for opt in parser._actions[0].option_strings if opt != "-h"]
        del parser._option_string_actions['-h']

        # Single-letter, single-dash and other counterparts for non ambiguous flags
        parser.add_argument('-f', '--f', '-firstname', *_flags_generator("firstname"),
                            dest="first_name", help=SUPPRESS, default=SUPPRESS)

        parser.add_argument('-b', '--b', '-building', *_flags_generator("building"),
                            dest="building", help=SUPPRESS, default=SUPPRESS)

        parser.add_argument('-l', '--l', '-login', *_flags_generator("login"),
                            dest="login_name", help=SUPPRESS, default=SUPPRESS)

        parser.add_argument('-d', '--d', '-department', *_flags_generator("department"),
                            dest="department", help=SUPPRESS, default=SUPPRESS)

        parser.add_argument('-m', '--m', '-mobile', *_flags_generator("mobile"),
                            dest="mobile_phone", help=SUPPRESS, default=SUPPRESS)

        parser.add_argument('-c', '--c', '-ccid', *_flags_generator("ccid"),
                            dest="ccid", help=SUPPRESS, default=SUPPRESS)

        parser.add_argument('-u', '--u', '-uid', *_flags_generator("uid"),
                            dest="uid", help=SUPPRESS, default=SUPPRESS)

        domain_mutexes.add_argument('-y', '--y', '-yellowpages', *_flags_generator("yellowpages"),
                                    dest="yellowpages", action="store_true", help=SUPPRESS, default=SUPPRESS)

        formatting_mutexes.add_argument('-a', '--a', '-all', *_flags_generator("all"),
                                        dest="all", action="store_true", help=SUPPRESS, default=SUPPRESS)

        formatting_mutexes.add_argument('-t', '--t', '-terse', *_flags_generator("terse"),
                                        dest="terse", action="append", help=SUPPRESS, default=SUPPRESS)

        # Single-dash flags for ambiguous single-letter flags
        parser.add_argument('-office', *_flags_generator("office"), dest="office", help=SUPPRESS, default=SUPPRESS)
        parser.add_argument('-orderby', *_flags_generator("orderby"), dest="orderby", action="append",
                            help=SUPPRESS, default=SUPPRESS)

        parser.add_argument('-email', *_flags_generator("email"), dest="email", help=SUPPRESS, default=SUPPRESS)
        domain_mutexes.add_argument('-external', *_flags_generator("external"), dest="external", action="store_true",
                                    help=SUPPRESS, default=SUPPRESS)

        parser.add_argument('-surname', *_flags_generator("surname"), dest="last_name", help=SUPPRESS, default=SUPPRESS)
        parser.add_argument('-section', *_flags_generator("section"), dest="section", help=SUPPRESS, default=SUPPRESS)

        parser.add_argument('-phone', *_flags_generator("phone"), dest="landline_phone", help=SUPPRESS,
                            default=SUPPRESS)
        parser.add_argument('-pobox', *_flags_generator("pobox"), dest="post_box", help=SUPPRESS, default=SUPPRESS)

        parser.add_argument('-gid', *_flags_generator("gid"), dest="gid", help=SUPPRESS, default=SUPPRESS)
        parser.add_argument('-group', *_flags_generator("group"), dest="group", help=SUPPRESS, default=SUPPRESS)

        parser.add_argument('-homedir', *_flags_generator("homedir"), dest="home_directory", help=SUPPRESS,
                            default=SUPPRESS)
        parser.add_argument('-help', *_flags_generator("help"), dest="help", action="help", help=SUPPRESS,
                            default=SUPPRESS)

        # Ambiguous single-letter flags have NO EFFECT: even when followed by a string, they are ignored
        # and the following string treated as a searchstring
        parser.add_argument('-o', action='store_true', help=SUPPRESS, default=SUPPRESS)  # office, orderby
        parser.add_argument('-e', action='store_true', help=SUPPRESS, default=SUPPRESS)  # email, external
        parser.add_argument('-s', action='store_true', help=SUPPRESS, default=SUPPRESS)  # surname, section
        parser.add_argument('-p', action='store_true', help=SUPPRESS, default=SUPPRESS)  # phone, pobox
        parser.add_argument('-g', action='store_true', help=SUPPRESS, default=SUPPRESS)  # gid, group
        parser.add_argument('-h', action='store_true', help=SUPPRESS, default=SUPPRESS)  # help, homedir

        # ######################################## #
        #                 PARSING                  #
        # ######################################## #

        # Gather the parameters from the command line as a dictionary
        parameters = vars(parser.parse_args(args=None if sys.argv[1:] else ['--help']))

        # Print warning for ambiguous flags - they are going to be ignored
        for flag, ambiguity in [
            ('o', "office, orderby"),
            ('e', "email, external"),
            ('s', "surname, section"),
            ('p', "phone, pobox"),
            ('g', "gid, group"),
            ('h', "help, homedir")
        ]:
            if parameters.pop(flag, False):
                print(f"Option {flag} is ambiguous ({ambiguity})")

        # Filter out non-search related flags for later use
        search_string = parameters.pop('search_string', None)
        externals = parameters.pop('external', None)
        yellowpages = parameters.pop('yellowpages', None)
        all_fields = parameters.pop('all', None)
        orderby = parameters.pop('orderby', [])
        terse = parameters.pop('terse', [])
        verbose = parameters.pop('verbose', False)

        # Identify the search domain
        domain = "users"
        if externals:
            domain = "externals"
        elif yellowpages:
            domain = "services"

        # Validate terse fields, if needed:
        for field in terse:
            if field not in _TERSE_ALIASES.keys():
                print(f'Error: Unrecognized -t(erse) option value ({field}), see phonebook --help for help.')
                return

        # Validate orderby fields, if needed:
        for field in orderby:
            if field not in _ORDERBY_ALIASES.keys():
                print(f'Error: Unrecognized -order(by) option value ({field}), see phonebook --help for help.')
                return

        # Instantiate a PhoneBook. Note: it can throw LDAPErrors in case of failure (i.e. network failures)
        pb = PhoneBook(domain=domain)

        # If terse is set, we don't want to merge entries by CCID anymore (that's the behavior of the old CLI)
        merge = len(terse) == 0

        # Build the query and search
        query_parameters = _build_query(parameters, search_string)
        matches = pb.arbitrary_search(parameters=query_parameters, raw=False, merge_by_ccid=merge, verbose=verbose)

        # Sort the matches
        if orderby:
            _sort_by_fields(matches, orderby)
        else:
            _sort_by_fields(matches, ['surname', 'firstname'])

        # Print the entries, properly rendered
        [print(match._render(terse=terse, all_fields=all_fields)) for match in matches]
        if all_fields and not yellowpages:
            print(dedent("""\
                            #-------------------------------------------------------------------------------
                            #Account St(atus): P(rimary), S(econdary), s(ervice), U(nknown)
                            #                  A(ctive),  D(isabled),  P(assword expired),  L(ocked out) """))
        elif terse:
            print("#-------------------------------------------------------------------------------\n"
                  "WARNING! If you built a script on top of the PhoneBook CLI, please consider \n"
                  "using the PhoneBook Python API instead.\n"
                  "#-------------------------------------------------------------------------------")

    except Exception as e:
        if verbose:
            traceback.print_exc()
        print("-" * 80)
        print("A fatal exception occurred!")
        print(e)
        print("-" * 80)
        sys.exit(1)


def _build_query(parameters: Dict[str, Any], search_string: str = None) -> Tuple[str, Any]:
    """
    Given the search_string and all the other explicit flags, build the correct query for the search.

    :param parameters: all the explicit fields given in the command line, like ``--firstname John``
    :param search_string: the standalone string given without specifying any flag, usually a login name or phone
        number.
    :return: a well-formed ``PhoneBook`` query.
    """
    query_parameters_list = []

    # A free search string is interpreted as follows:
    #    - If it's numeric, look for a phone number (either long or short form)
    #    - If it isn't, look for a display name or account name
    if search_string:
        if search_string.strip().strip("+").isdigit():  # Phone numbers might begin with +, so check that too
            search_string_parameters = (PHONEBOOK_OR, [
                ("mobile", f"*{PhoneBookEntry.short_phone_number_to_long(search_string)}*"),
                ("mobile", f"*{PhoneBookEntry.long_phone_number_to_short(search_string)}*"),
                ("telephoneNumber", f"*{PhoneBookEntry.short_phone_number_to_long(search_string)}*"),
                ("telephoneNumber", f"*{PhoneBookEntry.long_phone_number_to_short(search_string)}*")
            ])
        else:
            search_string_parameters = (PHONEBOOK_OR, [
                ("displayName", f"*{search_string}*"),
                ("sAMAccountName", search_string)
            ])
        query_parameters_list.append(search_string_parameters)

    # Append all the other parameters as a list of tuples
    query_parameters_list.extend(parameters.items())

    # AND all the parameters
    complete_query = (PHONEBOOK_AND, query_parameters_list)

    return complete_query


def _sort_by_fields(matches: List[PhoneBookEntry], fields: List[str]) -> List[PhoneBookEntry]:
    """
    Mimick the old phonebook --orderby flag behavior. Orders the list of entries by the specified fields.
    :param matches: list of PhoneBook entries
    :param fields: Fields to sort the above list by
    :return: an ordered list of PhoneBookEntry objects
    """
    # Double check: technically the fields have already been validated
    if any(field not in _ORDERBY_ALIASES.keys() for field in fields):
        raise ValueError(f"Valid fields for orderby are: {_ORDERBY_ALIASES.keys()}")

    for field in reversed(fields):
        matches.sort(key=lambda entry: vars(entry)[_ORDERBY_ALIASES[field]][0].lower() if len(vars(entry)[_ORDERBY_ALIASES[field]]) > 0 else "-")
    return matches


def _flags_generator(flag: str):
    """
    The phonebook does not have a list of flags, but simply tries to pattern-match the incoming flags and
    guess which flag it is. argparse instead wants a fixed list of flags. This function gives to argparse
    all the flags that were supposed to match the regex in the Perl version.

    :param flag: the name of the flag, without dashed, like ``office``
    :return: list of all possible substrings, with one and two dashes. Single letter and full version excluded.
     Example: [-of, --of, -off, --off, ...]. Note that these 4 flags are MISSING: -o, --o, -office, --office.
     This happens because long flags will be separately shown in the help, which short strings might be ambiguous and
     are also handled separately.
    """
    flags_list = []
    for i in range(2, len(flag)):
        flags_list.append(f"-{flag[:i]}")
        flags_list.append(f"--{flag[:i]}")
    return flags_list


if __name__ == "__main__":
    _cli()
