from typing import List, Mapping
import re
import ast
import pytest
from pyphonebook import PhoneBook


@pytest.fixture
def phonebook(monkeypatch):
    pb = PhoneBook()
    pb._raw_search = fake_raw_search
    yield pb


@pytest.fixture
def mock_users(monkeypatch):
    yield CANDIDATES


# Fake but realistic return value for LDAP queries
FAKE_ENTRY_RAW_DATA_1 = (
    'CN=user1,OU=Users,OU=Organic Units,DC=cern,DC=ch',
    {
        'objectClass': [b'top',
                        b'person',
                        b'organizationalPerson',
                        b'user'],
        'cn': [b'user1'],
        'description': [b'CERN - BE/BI'],
        'postOfficeBox': [b'Z00010'],
        'physicalDeliveryOfficeName': [b'*Room*999 1-A11*'],
        'distinguishedName': [b'CN=user1,OU=Users,OU=Organic Units,DC=cern,DC=ch'],
        'instanceType': [b'4'],
        'whenCreated': [b'20180101120000.0Z'],
        'whenChanged': [b'20181231235959.0Z'],
        'displayName': [b'User One'],
        'uSNCreated': [b'1234567'],
        'memberOf': [b'CN=egroup-1,OU=e-groups,OU=Workgroups,DC=cern,DC=ch',
                     b'CN=egroup-2,OU=e-groups,OU=Workgroups,DC=cern,DC=ch',],
        'uSNChanged': [b'87654321'],
        'proxyAddresses': [b'SMTP:user.one@cern.ch',
                           b'smtp:user1@mail.cern.ch',
                           b'smtp:user1@cern.ch', b'sip:user.one@cern.ch'],
        'name': [b'user1'],
        'objectGUID': [b'\xa7\x84\xce\xfb)f\x89K\xa2&\xa2a\xd7\xe0j\xa3'],
        'objectSid': [b'\x01\x05\x00\x00\x1e\xb0#?\x0c\xccXe\xd6\xd3\x1c\xf7\x8a\xe5\x12L\xb8\x83\xc3Sl\xd8T~'],
        'managedObjects': [b'CN=device1,OU=CERN Linux Computers,DC=cern,DC=ch'],
        'userPrincipalName': [b'user.one@cern.ch'],
        'objectCategory': [b'CN=Person,CN=Schema,CN=Configuration,CN={D1EC1E39-8235-4943-8A74-D0304B20B65C}'],
        'dSCorePropagationData': [b'14010101000000.0Z'],
        'msExchHideFromAddressLists': [b'FALSE'],
        'mailNickname': [b'user1'],
        'cernGroup': [b'BI'],
        'altSecurityIdentities': [
          b'X509:<I>DC=ch,DC=cern,CN=CERN Certification Authority<S>DC=ch,DC=cern,OU=Organic Units,OU=Users,CN=lsfcert',
          b'X509:<I>DC=ch,DC=cern,CN=CERN Certification Authority<S>DC=ch,DC=cern,OU=Organic Units,OU=Users,CN=acronmc'
        ],
        'gidNumber': [b'1234'],
        'cernMailType': [b'CERN Mailbox'],
        'directReports': [b'CN=pcbe12345,OU=CERN Linux Computers,DC=cern,DC=ch',
                          b'CN=device1,OU=CERN Linux Computers,DC=cern,DC=ch'],
        'extensionAttribute12': [b'SW'],
        'extensionAttribute14': [b'Monday 01 Jan 2020 at 12:00 AM'],
        'kopanoUserServer': [b'virtual-kopano'],
        'extensionAttribute13': [b'BI'],
        'preferredLanguage': [b'EN'],
        'department': [b'BE/BI'],
        'legacyExchangeDN': [b'/o=cern/ou=exchange administrative group (fydibohf23spdlt)/cn=recipients/cn=user1243'],
        'givenName': [b'User'],
        'kopanoAccount': [b'1'],
        'uidNumber': [b'123456'],
        'mail': [b'user.one@cern.ch'],
        'manager': [b'CN=boss,OU=Users,OU=Organic Units,DC=cern,DC=ch'],
        'company': [b'Not CERN'],
        'unixHomeDirectory': [b'/afs/cern.ch/user/u/user1'],
        'msExchHomeServerName': [
            b'/o=CERN/ou=Exchange Administrative Group (FYDIBOHF23SPDLT)/cn=Configuration/cn=Servers/cn=cernxchg43'],
        'loginShell': [b'/bin/bash'],
        'cernExternalMail': [b'personalemail1@gmail.com'],
        'cernAccountType': [b'Primary'],
        'extensionAttribute10': [b'personalemail1@gmail.com'],
        'cernInstituteName': [b'Not CERN'],
        'sn': [b'One'],
        'employeeID': [b'654321'],
        'employeeType': [b'Primary'],
        'division': [b'BE'],
        'cernSection': [b'SW'],
        'sAMAccountName': [b'user1'],
        'userAccountControl': [b'321'],
        'gecos': [b'User One,999 1-A11,,'],
        'extensionAttribute9': [b'CERN Mailbox']
    }
)
FAKE_ENTRY_RAW_DATA_2 = (
    'CN=user2,OU=Users,OU=Organic Units,DC=cern,DC=ch',
    {
        'objectClass': [b'top',
                        b'person',
                        b'organizationalPerson',
                        b'user'],
        'cn': [b'user2'],
        'description': [b'CERN - BE/BI'],
        'postOfficeBox': [b'Z00000'],
        'physicalDeliveryOfficeName': [b'*Room*999 2-A22*'],
        'distinguishedName': [b'CN=user2,OU=Users,OU=Organic Units,DC=cern,DC=ch'],
        'instanceType': [b'4'],
        'whenCreated': [b'20180101120000.0Z'],
        'whenChanged': [b'20181231235959.0Z'],
        'displayName': [b'User Two'],
        'uSNCreated': [b'2234567'],
        'memberOf': [b'CN=egroup-2,OU=e-groups,OU=Workgroups,DC=cern,DC=ch',
                     b'CN=egroup-3,OU=e-groups,OU=Workgroups,DC=cern,DC=ch',],
        'uSNChanged': [b'87654322'],
        'proxyAddresses': [b'SMTP:user.two@cern.ch',
                           b'smtp:user2@mail.cern.ch',
                           b'smtp:user2@cern.ch', b'sip:user.two@cern.ch'],
        'name': [b'user2'],
        'objectGUID': [b'\xa7\x84\xce\xfb)f\x89K\xa2&\xa2a\xd7\xe0j\xa3'],
        'objectSid': [b'\x01\x05\x00\x00\x1e\xb0#?\x0c\xccXe\xd6\xd3\x1c\xf7\x8a\xe5\x12L\xb8\x83\xc3Sl\xd8T~'],
        'managedObjects': [b'CN=device9,OU=CERN Linux Computers,DC=cern,DC=ch'],
        'userPrincipalName': [b'user.two@cern.ch'],
        'objectCategory': [b'CN=Person,CN=Schema,CN=Configuration,CN={D1EC1E39-8235-4953-8A94-D0304D22B65C}'],
        'dSCorePropagationData': [b'15010101000000.0Z'],
        'msExchHideFromAddressLists': [b'FALSE'],
        'mailNickname': [b'user2'],
        'cernGroup': [b'CO'],
        'altSecurityIdentities': [
          b'X509:<I>DC=ch,DC=cern,CN=CERN Certification Authority<S>DC=ch,DC=cern,OU=Organic Units,OU=Users,CN=lsfcert',
          b'X509:<I>DC=ch,DC=cern,CN=CERN Certification Authority<S>DC=ch,DC=cern,OU=Organic Units,OU=Users,CN=acronmc'
        ],
        'gidNumber': [b'2234'],
        'cernMailType': [b'CERN Mailbox'],
        'directReports': [b'CN=pcbe22345,OU=CERN Linux Computers,DC=cern,DC=ch',
                          b'CN=device9,OU=CERN Linux Computers,DC=cern,DC=ch'],
        'extensionAttribute12': [b'APS'],
        'extensionAttribute14': [b'Monday 02 Jan 2020 at 11:00 AM'],
        'kopanoUserServer': [b'virtual-kopano'],
        'extensionAttribute13': [b'CO'],
        'preferredLanguage': [b'FR'],
        'department': [b'BE/CO'],
        'legacyExchangeDN': [b'/o=cern/ou=exchange administrative group (fydibohf23spdlt)/cn=recipients/cn=user2243'],
        'givenName': [b'User'],
        'kopanoAccount': [b'1'],
        'uidNumber': [b'223456'],
        'mail': [b'user.two@cern.ch'],
        'manager': [b'CN=boss,OU=Users,OU=Organic Units,DC=cern,DC=ch'],
        'company': [b'CERN'],
        'unixHomeDirectory': [b'/afs/cern.ch/user/u/user2'],
        'msExchHomeServerName': [
            b'/o=CERN/ou=Exchange Administrative Group (FYDIBOHF23SPDLT)/cn=Configuration/cn=Servers/cn=cernxchg43'],
        'loginShell': [b'/bin/bash'],
        'cernExternalMail': [b'personalemail2@gmail.com'],
        'cernAccountType': [b'Primary'],
        'extensionAttribute10': [b'personalemail2@gmail.com'],
        'cernInstituteName': [b'CERN'],
        'sn': [b'Two'],
        'employeeID': [b'654382'],
        'employeeType': [b'Primary'],
        'division': [b'BE'],
        'cernSection': [b'APS'],
        'sAMAccountName': [b'user2'],
        'userAccountControl': [b'322'],
        'gecos': [b'User Two,999 2-A22,,'],
        'extensionAttribute9': [b'CERN Mailbox']
    }
)
FAKE_ENTRY_RAW_DATA_3 = (
    'CN=user3,OU=Users,OU=Organic Units,DC=cern,DC=ch',
    {
        'objectClass': [b'top',
                        b'person',
                        b'organizationalPerson',
                        b'user'],
        'cn': [b'user3'],
        'description': [b'CERN - EP/DT'],
        'postOfficeBox': [b'Z06300'],
        'physicalDeliveryOfficeName': [b'Bld.*999 *'],
        'distinguishedName': [b'CN=user3,OU=Users,OU=Organic Units,DC=cern,DC=ch'],
        'instanceType': [b'4'],
        'whenCreated': [b'20180101130000.0Z'],
        'whenChanged': [b'20181231135959.0Z'],
        'displayName': [b'User Three'],
        'uSNCreated': [b'2374567'],
        'memberOf': [b'CN=egroup-5,OU=e-groups,OU=Workgroups,DC=cern,DC=ch',
                     b'CN=egroup-6,OU=e-groups,OU=Workgroups,DC=cern,DC=ch',],
        'uSNChanged': [b'87674322'],
        'proxyAddresses': [b'SMTP:user.three@cern.ch',
                           b'smtp:user32@mail.cern.ch',
                           b'smtp:user3@cern.ch', b'sip:user.three@cern.ch'],
        'name': [b'user3'],
        'objectGUID': [b'\xa7\x84\xce\xfb)f\x89K\xa2&\xa2a\xd7\xe0j\xa3'],
        'objectSid': [b'\x01\x05\x00\x00\x1e\xb0#?\x0c\xccXe\xd6\xd3\x1c\xf7\x8a\xe5\x12L\xb8\x83\xc3Sl\xd8T~'],
        'managedObjects': [b'CN=device8,OU=CERN Linux Computers,DC=cern,DC=ch'],
        'userPrincipalName': [b'user.three@cern.ch'],
        'objectCategory': [b'CN=Person,CN=Schema,CN=Configuration,CN={D1EC1E39-8235-4953-8A94-D0304D22B65C}'],
        'dSCorePropagationData': [b'15050101000000.0Z'],
        'msExchHideFromAddressLists': [b'FALSE'],
        'mailNickname': [b'user3'],
        'cernGroup': [b'DT'],
        'altSecurityIdentities': [
          b'X509:<I>DC=ch,DC=cern,CN=CERN Certification Authority<S>DC=ch,DC=cern,OU=Organic Units,OU=Users,CN=lsfcert',
          b'X509:<I>DC=ch,DC=cern,CN=CERN Certification Authority<S>DC=ch,DC=cern,OU=Organic Units,OU=Users,CN=acronmc'
        ],
        'gidNumber': [b'2634'],
        'cernMailType': [b'CERN Mailbox'],
        'directReports': [b'CN=pcep22775,OU=CERN Linux Computers,DC=cern,DC=ch',
                          b'CN=device8,OU=CERN Linux Computers,DC=cern,DC=ch'],
        'extensionAttribute14': [b'Monday 02 Jan 2020 at 11:00 AM'],
        'kopanoUserServer': [b'virtual-kopano'],
        'extensionAttribute13': [b'DT'],
        'preferredLanguage': [b'EN'],
        'department': [b'EP/DT'],
        'legacyExchangeDN': [b'/o=cern/ou=exchange administrative group (fydibohf23spdlt)/cn=recipients/cn=user2743'],
        'givenName': [b'User'],
        'kopanoAccount': [b'1'],
        'uidNumber': [b'273456'],
        'mail': [b'user.three@cern.ch'],
        'manager': [b'CN=otherboss,OU=Users,OU=Organic Units,DC=cern,DC=ch'],
        'company': [b'CERN'],
        'unixHomeDirectory': [b'/afs/cern.ch/user/u/user3'],
        'msExchHomeServerName': [
            b'/o=CERN/ou=Exchange Administrative Group (FYDIBOHF23SPDLT)/cn=Configuration/cn=Servers/cn=cernxchg43'],
        'loginShell': [b'/bin/bash'],
        'cernExternalMail': [b'personalemail3@gmail.com'],
        'cernAccountType': [b'Primary'],
        'extensionAttribute10': [b'personalemail3@gmail.com'],
        'cernInstituteName': [b'CERN'],
        'sn': [b'Three'],
        'employeeID': [b'654322'],
        'employeeType': [b'Primary'],
        'division': [b'EP'],
        'sAMAccountName': [b'user3'],
        'userAccountControl': [b'352'],
        'gecos': [b'User Three,999 3-A33,,'],
        'extensionAttribute9': [b'CERN Mailbox'],
        'facsimileTelephoneNumber': [b'*1234567'],
        'telephoneNumber': [b'+41227697554']
    }
)
CANDIDATES = (FAKE_ENTRY_RAW_DATA_1, FAKE_ENTRY_RAW_DATA_2, FAKE_ENTRY_RAW_DATA_3)


def fake_raw_search(query):
    return candidates_matching_query(parse_fake_query(query))


def candidates_matching_query(query):
    matches = []
    # The statements are evaluated per user
    for id, candidate in enumerate(CANDIDATES):
        if user_matches_query(query, candidate[1]):
            matches.append(candidate)
    return matches


def parse_fake_query(query):
    print("Raw query: ", query)

    # places all the words between double quotes
    # and appends a comma after each
    query = re.sub(r'([^()]+)', r'"\1",', query)

    # Convert into list brackets and correct commas
    query = query.replace("(", "[") \
                 .replace(")", "]") \
                 .replace("][", "], [") \
                 .replace("\",]", "\"]")

    # safely evaluates the resulting string
    print("Query to eval: ", query)
    return ast.literal_eval(query)


def user_matches_query(query, candidate):
    print("Matching user ", candidate["cn"][0], " against query: ", query)

    # AND query
    if query[0] == "&":
        print("AND query found")
        result = True  # In case of AND, adding a True does not change the statement's truth value
        for statement in query[1:]:
            result = result and user_matches_query(statement, candidate)  # Iterate down on each statement
        return result

    # OR query
    elif query[0] == "|":
        print("OR query found")
        result = False  # In case of OR, adding a False does not change the statement's truth value
        for statement in query[1:]:
            result = result or user_matches_query(statement, candidate)  # Iterate down on each statement
        return result

    # Simple tuple value - and recursion stopping condition
    else:
        print("Simple statement found")
        search_key, search_value = query[0].split("=")
        return user_matches_statement(search_key, search_value, candidate)


def user_matches_statement(search_key, search_value, user_values):
    """
        Checks if a user contains the specified parameter,
        as a search_key=search_value statement.
        Candidate values are always in the form [b'value']
        So:
           'value1' == [b'value1']  ->  True
           'value1' == [b'value2']  ->  False
           'value1' == []  ->  False
           '' == []  ->  True
           None == None  ->  True
           'value1' == [b'value1', b'value2']  ->  True
        """
    if search_key in user_values.keys():
        candidate_values = user_values[search_key]

        # Match null-like values
        if (candidate_values is None or candidate_values == []) and (search_value is None or search_value == ""):
            return True

        for candidate_value in candidate_values:
            candidate_value = candidate_value.decode()

            # Found matching value
            if search_value == candidate_value:
                return True

            # Get around star search
            elif search_value.find("*") > -1 and search_value.replace("*", "").strip() in candidate_value:
                return True

        # No match
        return False




